#!/bin/bash

# Author: Victor Correal (victor.correal@bsc.es)
# Date  : 28/06/2024

UTILS_DIR="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"
source ${UTILS_DIR}/bash_utils.sh

usage(){
    info "Usage: run_nemo.sh "
    info "\t--nemo   <NEMO_CFG_PATH>"
    info "\t--xios   <XIOS_BIN>"
    info "\t--env    <ENV_FILE"
    info "\t--rundir <RUNDIR_PATH>"
    info "\t--resolution  ORCA025|ORCA12|ORCA36"
    info "OPTIONAL FLAGS:"
    info "[--nodes=<nodes>] [--MPI=<val>] [--OpenMP=<val>] "
    info "[--timelimit=<hh:mm:ss>]"
    info "[--account=<SLURM-ACC>] [--queue=<queue>] [--wait] [--contiguous]"
    info "[--force] [--dry] [--help] [--highmem]"
}

parse_args(){
    OPTIONS=$(getopt -o "" --long highmem,dry,nemo:,xios:,env:,rundir:,resolution:,nodes:,account:,queue:,MPI:,OpenMP:,timelimit:,force,help,wait,contiguous -n "$0" -- "$@")

    eval set -- "$OPTIONS"

    while true; do
        case "$1" in
        --highmem)
            SLURM_HIGHMEM=true
            shift
        ;;
        --nemo)
            NEMO="$2"
            shift 2
        ;;
        --xios)
            XIOS_BIN="$2"
            shift 2
        ;;
        --env)
            ENV_FILE="$2"
            shift 2
        ;;
        --rundir)
            RUNDIR="$2"
            shift 2
        ;;
        --resolution)
            RESOLUTION="$2"
            shift 2
        ;;
        --nodes)
            NODES="$2"
            shift 2
        ;;
        --account)
            ACCOUNT="$2"
            shift 2
        ;;
        --queue)
            QUEUE="$2"
            shift 2
        ;;
        --MPI)
            MPI_PER_NODE="$2"
            shift 2
        ;;
        --OpenMP)
            OMP_PER_TASK="$2"
            shift 2
        ;;
        --timelimit)
            TIME_LIMIT="$2"
            shift 2
        ;;
        --help)
            usage
            exit 0
        ;;
        --force)
            FORCE_RUN=true
            shift
        ;;
        --wait)
            SLURM_WAIT=true
            shift
        ;;
        --contiguous)
            SLURM_CONTIGUOUS=true
            shift
        ;;
        --dry)
            DRY=true
            shift
        ;;
        --)
            shift
            break
        ;;
        *)
            usage
            exit 1
        ;;
        esac
    done
}


check_required_args(){
    [ -z $NEMO ] && fatal "NEMO binary not set, use --nemo"
    [ -z $XIOS_BIN ] && fatal "XIOS binary not set, use --xios"
    [ -z $ENV_FILE ] && fatal "ENV file not set use --env"
    [ -z $RUNDIR ] && fatal "RUNDIR not set, use --rundir"
    [ -z $RESOLUTION ] &&  fatal "RESOLUTION not set, use --resolution"
}

check_dirs_and_bins(){
    [ ! -d $NEMO ] && fatal "Unable to find bin file at $BIN_FILE"
    [ ! -f $XIOS_BIN ] && fatal "Unable to find bin file at $XIOS_BIN"
    [ ! -f $ENV_FILE ] && fatal "Unable to find env file at $ENV_FILE"
    #[ -d $RUNDIR ] && fatal "Rundir dir already exists"
}

check_slurm_options(){
    if [ "$ACCOUNT" != "DEFAULT" ]; then
        SLURM_ARGS+="-A $ACCOUNT "
        info "Overriding default for SLURM ACCOUNT: $ACCOUNT"
    fi
    if [ "$QUEUE" != "DEFAULT" ]; then
        SLURM_ARGS+="--qos $QUEUE "
        info "Overiding default for QUEUE: $QUEUE"
    fi
    if [ "$NODES" != "DEFAULT" ]; then
        SLURM_ARGS+="-N $NODES "
        info "Overiding default for NODES: $NODES"
    fi
    if [ "$TIME_LIMIT" != "DEFAULT" ]; then
        SLURM_ARGS+="--time $TIME_LIMIT "
        info "Overiding default for TIME_LIMIT: $TIME_LIMIT"
    fi

    if [ "$MPI_PER_NODE" != "DEFAULT" ]; then
        SLURM_ARGS+="--ntasks-per-node $MPI_PER_NODE "
        info "Overiding default for MPI_PER_NODE: $MPI_PER_NODE"
    fi

    if [ "$OMP_PER_TASK" != "DEFAULT" ]; then
        SLURM_ARGS+="--cpus-per-task $OMP_PER_TASK "
        info "Overiding default for OMP_PER_TASK: $OMP_PER_TASK"
    fi
    if [ $SLURM_WAIT = true ]; then
        SLURM_ARGS+="-W "
        info "Waiting sbatch to finish the job before exit sbatch!"
    fi
    if [ $SLURM_CONTIGUOUS = true ]; then
        SLURM_ARGS+="--contiguous "
        info "Using SLURM contiguous"
    fi
    if [ $SLURM_HIGHMEM = true ]; then
        SLURM_ARGS+="-C highmem "
        info "Using SLURM highmem"    
    fi
    SLURM_ARGS+=" --exclusive -C perfparanoid"
}


print_options(){
    info "=============================MODEL============================="
    info "NEMO CFG: $NEMO"
    info "XIOS: $XIOS_BIN"
    info "ENV: $ENV_FILE"
    info "RUNDIR: $RUNDIR"
    info "==========================JOB OPTIONS=========================="
    info "RESOLUTION: $RESOLUTION"
    [ $SLURM_WAIT = true ] && info "SLURM_WAIT: ON"
    [ $SLURM_CONTIGUOUS = true ] && info "SLURM_CONTIGUOS: ON"
    info "====================ALLOCATION RESOURCES======================="
    info "$SLURM_ARGS"
    info "==============================================================="
}

check_user(){
    if [ $FORCE_RUN = false ]; then
        echo "FINE? (Y/N)"
        read ANSW
        case "$ANSW" in
            [Yy])
                success "OK!"
                ;;
            *)
                fatal "USER INPUT $ANSW"
                ;;
        esac

    else
        echo "FORCE RUN enable"
    fi
}

change_namelist_parameter() {
    local namelist="$1"
    local parameter="$2"
    local new_val="$3"

    info "*** $namelist updating parameter $parameter -> $new_val"

    # Usar sed para buscar el parámetro y reemplazar su valor
    sed -i -E "/^\s*&/,/^\s*\// s/^(\s*$parameter\s*=\s*)[^,]*$/\1$new_val/" "$namelist"
}
